#!/usr/bin/perl
use 5.010;
use strict;
use warnings;
use Time::HiRes;


# Globals
our %sym_tab;			# Symbol Table
our $progname;			# base program name (no extension)
our $binfh;			# Low byte hex file handle


print "STUPID Computer Assembler V0.00\n";

if ($#ARGV<0) {
	print "Usage: $0 <src_file>\n";
	exit 1;
}

$progname = $ARGV[0];if ($progname =~ '.') {
	$progname = substr($ARGV[0],0,rindex($ARGV[0],'.'));
}

pass1();
pass2();

print "$ARGV[0] assembled successfully\n";

exit(0);

# Pass 1: Build Symbol Table
sub pass1 {
	print "==Pass 1==\n";
	my $start = Time::HiRes::gettimeofday();
	my $num_statements = 0;
	my $addr = 0;
	my $linenum = 0;
	my $line;
	my %reserved = (
	"ac"  => 0,
	"ir0" => 0,
	"ir1" => 0,
	"ir2" => 0,
	"ir3" => 0,
	"pr"  => 0,
	);

	# Open the input file
	open(my $in, "<", $ARGV[0]) or die "Can't open $ARGV[0]: $!";
	
	while (<$in>) {
		$linenum++;
		# Strip off comments & leading/trailing white space
		my $comment_start = index($_,";");
		
		if ($comment_start < 0) {
			$comment_start = length;
		}
		$line = trim(substr($_,0,$comment_start));
		# Ignore blank lines
		next if $line eq "";

		# Process directives
		if (substr($line,0,4) eq ".end") {
			last;
		}
		if (substr($line,0,4) eq ".loc") {
			$addr = hex(trim(substr($line,4)));
			
			if (($addr < 0) || ($addr > 0xFFF)) {
				printf("Line $linenum: Invalid address specified: %X\n",$addr);
				die;
			}
			next;
		}
		
		# Process labels
		my $cend = index($line,':');
		if ($cend > 0) {
			# This line contains a label
			my $label = substr($line,0,$cend);
			# Save label and its address in symbol table
			# First check label for validity
			die "Line $linenum: Invalid label: $label\n" unless $label =~ /^[_\$a-z][_\$a-z0-9]*$/;
			die "Line $linenum: Reserved label: $label\n" if exists $reserved{$label};

			# Now check for duplicate label
			die "Line $linenum: Duplicate label: $label\n" if exists $main::sym_tab{$label};
			$main::sym_tab{$label} = $addr;

			# Strip the label and process the rest of the line
			$line = trim(substr($line,$cend+1));
			next if $line eq "";
		} elsif ($cend == 0) {
			die "Line $linenum: Syntax error:\n$_\n";
		}
		
		if (substr($line,0,4) eq ".txt") {
			# Determine size of string & adjust $addr
			$line = trim(substr($line,4));
			# String should begin and end with "
			die "Line $linenum: missing \" in .txt string\n" unless $line =~ /^\".*\"$/;
			$line = substr($line,1,-1);
			$line =~ s/\\//g;
			$addr += length($line) + 1;
			next;
		}
		if (substr($line,0,4) eq ".blk") {
			$line = trim(substr($line,4));
			if ($line =~ /^\d/) {
				if ($line =~ /^0/) {
					$addr += oct $line;
				} else {
					$addr += $line;
				}
			}
			next;
		}

		# Process statements and constants.  Assume that we have a statement or numeric
		# value at this point - we'll catch syntax errors in pass 2... 
		$addr++;
		die "Line $linenum: Program exceeds 4K\n" if $addr > 0x1000;
		$num_statements++;
	}
	

	my $numsym = keys %main::sym_tab;
	my $end = Time::HiRes::gettimeofday();
	printf "Elapsed time: %.3f\n", $end-$start;
	print "Program contains $numsym symbol(s), $num_statements statement(s).\n\n";
	close $in or die "$in: $!";
}



# Pass 2: Generate binary file and listing file
sub pass2 {
	my $start = Time::HiRes::gettimeofday();
	my $addr = 0;
	my $linenum = 0;
	my $line;
	my $token;
	my $cend;
	my $instr;
	my $isbranch;
	my $offset;
	my $i;
	my $c;
	my $first;
	my %instr0 = (			# Instruction that require no arguments
	"asl"    => 0xFEB,		# Arithmetic Shift Left (AC)
	"asr"    => 0xFE8,		# Arithmetic Shift Right (AC)
	"break"  => 0xFE3,		# Breakpoint
	"clr"    => 0xFF1,		# Clear AC
	"clrc"   => 0xFFA,		# Clear Carry
	"com"    => 0xFF3,		# Complement AC
	"dec"    => 0xFF6,		# Decrement AC
	"getc"   => 0xFE1,		# Read character from TTY (to AC)
	"halt"   => 0xFFF,		# Halt Processor
	"inc"    => 0xFF5,		# Increment AC
	"neg"    => 0xFF4,		# Negate AC
	"nop"    => 0xFF0,		# No op
	"putc"   => 0xFE2,		# Send character to TTY (from AC)
	"readsw" => 0xFE0,		# Read front panel switches (to AC)
	"rol"    => 0xFEC,		# Rotate AC left (no carry)
	"rolc"   => 0xFED,		# Rotate AC left (through carry)
	"ror"    => 0xFE9,		# Rotate AC right (no carry)
	"rorc"   => 0xFEA,		# Rotate AC right (through carry)
	"scc"    => 0xFEF,		# Skip if carry clear
	"scs"    => 0xFEE,		# Skip if carry set
	"set"    => 0xFF2,		# Set AC to all ones
	"setc"   => 0xFF9,		# Set carry
	"sicr"   => 0xFE4,		# Skip if input character received
	"sobe"   => 0xFE5,		# Skip if output buffer empty
	);
	my %instrpr = (			# Page register instruction
	"setpr"    => 0xFD0,		# Set page register
	);
	my %instrac = (			# Instructions that require AC as target
	"sneg"   => 0xFFE,		# Skip if negative
	"spos"   => 0xFFD,		# Skip if positive
	);
	my %instrir = (			# Instructions that require IR reg as target
	"jmp"    => 0xFC0,		# Jump to (IRx)
	"ret"    => 0xFC0,		# Return to (IRx) (Same as jmp)
	);
	my %instrtgt = (		# Instructions that reference memory
	"add"    => 0x400,		# Add to AC
	"addc"   => 0x500,		# Add w/carry to AC
	"and"    => 0x800,		# Logical AND w/AC
	"br"     => 0x000,		# Relative Branch
	"lda"    => 0x200,		# Load value to AC
	"or"     => 0x900,		# Logical OR w/AC
	"sta"    => 0x300,		# Store value from AC
	"sub"    => 0x600,		# Subtract from AC
	"subb"   => 0x700,		# Subtract w/borrow from AC
	"swap"   => 0xE00,		# Swap AC & mem location values
	"xor"    => 0xA00,		# Logical XOR w/AC
	);
	my %instrtgta = (		# Instructions that reference memory or AC
	"dsz"    => 0xC00,		# Decrement and skip if result is zero
	"isz"    => 0xB00,		# Increment and skip if result is zero
	);
	my %instrtgtai = (		# Instruction that references mem, AC, or Indirect
	"sz"     => 0xD00,		# Skip if zero
	);
	my %instracir = (		# Instruction that references AC or indirect
	"snz"    => 0xFFC,		# Skip if non-zero
	);
	my %instrcall = (		# call instruction
	"call"   => 0xFC0,		# call via indirect reg & save retaddr
	);
	my %instrmov = (		# mov instruction
	"mov"    => 0xF00,		# move between registers
	);
	my %jrtgt = (			# jmp/ret target
	"(ir0)"  => 0x000,
	"(ir1)"  => 0x005,
	"(ir2)"  => 0x00A,
	"(ir3)"  => 0x00F,
	);
	my %snztgt = (			# snz target
	"ac"     => 0xFFC,
	"(ir0)"  => 0xFB2,
	"(ir1)"  => 0xFB7,
	"(ir2)"  => 0xFB9,
	"(ir3)"  => 0xFBD,
	);
	my %sztgt = (			# sz target
	"(ir0)"  => 0xFB1,
	"(ir1)"  => 0xFB6,
	"(ir2)"  => 0xFB8,
	"(ir3)"  => 0xFBC,
	);
	my %calltgt = (			# call target
	"(ir0),ir1" => 0xFC1,
	"(ir0),ir2" => 0xFC2,
	"(ir0),ir3" => 0xFC3,
	"(ir1),ir0" => 0xFC4,
	"(ir1),ir2" => 0xFC6,
	"(ir1),ir3" => 0xFC7,
	"(ir2),ir0" => 0xFC8,
	"(ir2),ir1" => 0xFC9,
	"(ir2),ir3" => 0xFCB,
	"(ir3),ir0" => 0xFCC,
	"(ir3),ir1" => 0xFCD,
	"(ir3),ir2" => 0xFCE,
	);
	my %movtgt = (			# mov target
	"ac,ir0"    => 0xF00,
	"ac,ir1"    => 0xF05,
	"ac,ir2"    => 0xF0A,
	"ac,ir3"    => 0xF0F,
	"ac,(ir0)"  => 0xF01,
	"ac,(ir1)"  => 0xF04,
	"ac,(ir2)"  => 0xF08,
	"ac,(ir3)"  => 0xF0C,
	"++ac,ir0"  => 0xF14,
	"++ac,ir1"  => 0xF19,
	"++ac,ir2"  => 0xF12,
	"++ac,ir3"  => 0xF13,
	"--ac,ir0"  => 0xF18,
	"--ac,ir1"  => 0xF1D,
	"--ac,ir2"  => 0xF1E,
	"--ac,ir3"  => 0xF17,
	"ir0,ac"    => 0xF02,
	"ir1,ac"    => 0xF06,
	"ir2,ac"    => 0xF09,
	"ir3,ac"    => 0xF0D,
	"++ir0,ac"  => 0xF11,
	"++ir1,ac"  => 0xF15,
	"++ir2,ac"  => 0xF1A,
	"++ir3,ac"  => 0xF1C,
	"--ir0,ac"  => 0xF10,
	"--ir1,ac"  => 0xF16,
	"--ir2,ac"  => 0xF1B,
	"--ir3,ac"  => 0xF1F,
	"(ir0),ac"  => 0xF03,
	"(ir1),ac"  => 0xF07,
	"(ir2),ac"  => 0xF0B,
	"(ir3),ac"  => 0xF0E,
	"ir0,ir1"   => 0xFA1,
	"ir0,ir2"   => 0xFA2,
	"ir0,ir3"   => 0xFA3,
	"ir1,ir0"   => 0xFA4,
	"ir1,ir2"   => 0xFA6,
	"ir1,ir3"   => 0xFA7,
	"ir2,ir0"   => 0xFA8,
	"ir2,ir1"   => 0xFA9,
	"ir2,ir3"   => 0xFAB,
	"ir3,ir0"   => 0xFAC,
	"ir3,ir1"   => 0xFAD,
	"ir3,ir2"   => 0xFAE,
	);
	my %escchar = (
	"0" => 000,	# null
	"r" => 015,	# cr
	"n" => 012,	# lf
	"t" => 011,	# tab
	"f" => 014,	# form feed
	"e" => 033,	# esc
	"b" => 010,	# backspace
	"a" => 007,	# alert (bel)
	);


	print "==Pass 2==\n";

	$instr = 0x000;			# DEBUG
	# Open the input file
	open(my $in, "<", $ARGV[0]) or die "Can't open $ARGV[0]: $!";
	# Open Listing file and binary file.
	my $listname = $progname . '.lst';
	open(my $listfh, ">", $listname) or die "Can't open listing file: $listname: $!";
	openbinfile();

	while (<$in>) {
		$linenum++;
		# Strip off comments & leading/trailing white space
		my $comment_start = index($_,";");
		
		if ($comment_start < 0) {
			$comment_start = length;
		}
		$line = trim(substr($_,0,$comment_start));
		# Ignore blank lines
		if ($line eq "") {
			printf {$listfh} "%04d            %s",$linenum,$_;
			next;
		}

		# Process directives
		if (substr($line,0,4) eq ".end") {
			printf {$listfh} "%04d            %s",$linenum,$_;
			# FixFixFix
			# Parse start addr
			$addr = 0;
			# Output END record
			writebin(0xE000 | $addr);
			last;
		}
		if (substr($line,0,4) eq ".loc") {
			# We already checked that the address was valid on the first pass
			$addr = hex(trim(substr($line,4)));
			printf {$listfh} "%04d %03X        %s",$linenum,$addr,$_;

			# Write address record to binary file
			writebin(0xA000 | $addr);
			next;
		}
		# Process labels (by ignoring them on this pass)
		my $cend = index($line,':');
		if ($cend > 0) {
			# This line contains a label
			# Strip the label and process the rest of the line
			$line = trim(substr($line,$cend+1));
			if ($line eq "") {
				printf {$listfh} "%04d            %s",$linenum,$_;
				next;
			}
		}

		if (substr($line,0,4) eq ".txt") {
			$first = 1;
			# Generate binary for null-terminated text string
			# As this is written, it is not necessary to escape quotes within
			# a string.
			$line = trim(substr($line,4));
			# Already verified that string begins and ends with "
			$line = substr($line,1,-1);
			# Make sure it doesn't end with \
			die "Line $linenum: missing close quote in .txt string\n" if $line =~ /\\$/;
			for ($i=0; $i<length($line); $i++) {
				$c = substr($line,$i,1);
				if  ($c eq '\\') {
					$c = substr($line,++$i,1);
					if (exists $escchar{$c}) {
						$c = chr($escchar{$c});
					}
				}
				if ($first) {
					printf {$listfh} "%04d %03X %03X    %s",$linenum,$addr,ord($c),$_;
					$first = "";
				} else {
					printf {$listfh} "     %03X %03X\n",$addr,ord($c);
				}
				writebin(ord($c));
				$addr++;
			}
			# Add trailing null
			$c = chr(0);
			if ($first) {
				printf {$listfh} "%04d %03X %03X    %s",$linenum,$addr,ord($c),$_;
				$first = "";
			} else {
				printf {$listfh} "     %03X %03X\n",$addr,ord($c);
			}
			writebin(0);
			$addr++;
			next;
		}
		if (substr($line,0,4) eq ".blk") {
			$token = trim(substr($line,4));
			printf {$listfh} "%04d %03X        %s",$linenum,$addr,$_;
			if ($token =~ /^\d/) {
				if ($token =~ /^0/) {
					$addr += oct $token;
				} else {
					$addr += $token;
				}
			}
			next;
		}

		# Get next token
		# Convert white space chars to space
		$line =~ s/\s+/ /g;
		$token = $line;
		my $zzz = index($line, ' ');
		if ($zzz > 0) {
			$token = substr($line,0,$zzz);
		}
		# Ignore blank lines
		if ($line eq "") {
			printf {$listfh} "%04d            %s",$linenum,$_;
			next;
		}

		# Check for numeric value or label
		if ($token =~ /^\d/) {
			if ($token =~ /^0/) {
				$instr = oct $token;
			} else {
				$instr = $token;
			}
		} elsif (exists $sym_tab{$token}) {
			$instr = $sym_tab{$token};
			# Did we gobble the whole line?
			die "Line $linenum: Syntax error:\n$_\n" unless length($token) eq length($line);
		}
		# Process statements
		elsif (exists $instr0{$token}) {
			# Instruction with no operands.  Easy case.
			$instr = $instr0{$token};
			die "Line $linenum: No target allowed.\n$_\n" unless length($token) eq length($line);
		} elsif (exists $instrpr{$token}) {
			# Instruction that operates on Page Register
			$instr = $instrpr{$token};
			$token = substr($line,length($token)+1);
			die "Line $linenum: Invalid page: $token" unless (length($token) eq 1);
			die "Line $linenum: Invalid page: $token" unless $token =~ /[0-9a-fA-F]/;
			$instr |= hex($token);
		} elsif (exists $instrac{$token}) {
			# Instruction that only operates on ac
			$instr = $instrac{$token};
			$token = substr($line,length($token)+1);
			die "Line $linenum: Invalid register: $token" unless ($token eq 'ac');
		} elsif (exists $instrir{$token}) {
			# Instructions that reference an indirect register
			$instr = $instrir{$token};
			die "Line $linenum: No jmp/ret target specified" if (length($token) eq length($line));
			$token = substr($line,length($token)+1);
			die "Line $linenum: Invalid jmp/ret target: $token" unless exists $jrtgt{$token};
			$instr |= $jrtgt{$token};
		} elsif (exists $instrtgt{$token}) {
			# Instructions that reference a memory location
			$instr = $instrtgt{$token};
			# Branch is sprecial case
			if ($token eq 'br') {
				$isbranch = 1;
			} else {
				$isbranch = "";
			}
			$token = substr($line,length($token)+1);
			# Target is either label, relative (./.+N/.-N) or absolute hex value
			# FixFixFix - This should be a subroutine called by instrtgt, instrtgta, instrtgtai!
			if ($token =~ /^\./) {
				$token =~ s/ //g;
				if ($token eq '.') {
					unless ($isbranch) {
						$instr |= ($addr % 256);
					}
				} elsif ($token =~ /^\.\+/) {
					$offset = substr($token,2);
					# FixFixFix - Verify decimal number
					if ($isbranch) {
						die "Line $linenum: Invalid branch address: $token" unless ($offset <=255);
						$instr |= $offset;
					} else {
						$instr |= ($addr + $offset) % 256;
					}
				} elsif ($token =~ /^\.\-/) {
					$offset = substr($token,2);
					# FixFixFix - Verify decimal number
					if ($isbranch) {
						die "Line $linenum: Invalid branch address: $token" unless ($offset <=255);
						$instr = 0x100 | $offset;
					} else {
						$instr |= ($addr - $offset) % 256;
					}
				} else {
					die "Line $linenum: Invalid relative address: $token";
				}
			} elsif (exists $sym_tab{$token}) {
				# A label!
				if ($isbranch) {
					$offset = $sym_tab{$token} - $addr;
					die "Line $linenum: Branch address out of range: $token" unless (abs($offset) <=255);
					if ($offset < 0) {
						$instr = 0x100 | abs($offset);
					} else {
						$instr |= $offset;
					}
				} else {
					$offset = ($sym_tab{$token} % 256);
					$instr |= $offset;
				}
			} else {
				# Hex or invalid
				# FixFixFix - interpret hex value
				die "Line $linenum: Invalid target address: $token";
			}

		} elsif (exists $instrtgta{$token}) {
			# Instructions that reference a mem loc. or ac (isz & dsz)
			$instr = $instrtgta{$token};
			$token = substr($line,length($token)+1);
			if ($token eq 'ac') {
				if ($instr eq 0xB00) {
					# isz ac => FF7
					$instr = 0xFF7;
				} else {
					# dsz ac => FF8
					$instr = 0xFF8;
				}
			} else {
				# Target is either label, relative (./.+N/.-N) or absolute hex value
				# FixFixFix - This should be a subroutine called by instrtgt, instrtgta, instrtgtai!
				if ($token =~ /^\./) {
					$token =~ s/ //g;
					if ($token eq '.') {
						$instr |= ($addr % 256);
					} elsif ($token =~ /^\.\+/) {
						$offset = substr($token,2);
						# FixFixFix - Verify decimal number
						$instr |= ($addr + $offset) % 256;
					} elsif ($token =~ /^\.\-/) {
						$offset = substr($token,2);
						# FixFixFix - Verify decimal number
						$instr |= ($addr - $offset) % 256;
					} else {
						die "Line $linenum: Invalid relative address: $token";
					}
				} elsif (exists $sym_tab{$token}) {
					# A label!
					$offset = ($sym_tab{$token} % 256);
					$instr |= $offset;
				} else {
					# Hex or invalid
					# FixFixFix - interpret hex value
					die "Line $linenum: Invalid target address: $token";
				}
			}
		} elsif (exists $instrtgtai{$token}) {
			# Instructions that reference a mem loc, ac, or indirect reg (sz)
			$instr = $instrtgtai{$token};
			$token = substr($line,length($token)+1);
			if ($token eq 'ac') {
				$instr = 0xFFB;
			} elsif (exists $sztgt{$token}) {
				$instr = $sztgt{$token};
			} else {
				# Target is either label, relative (./.+N/.-N) or absolute hex value
				# FixFixFix - This should be a subroutine called by instrtgt, instrtgta, instrtgtai!
				if ($token =~ /^\./) {
					$token =~ s/ //g;
					if ($token eq '.') {
						$instr |= ($addr % 256);
					} elsif ($token =~ /^\.\+/) {
						$offset = substr($token,2);
						# FixFixFix - Verify decimal number
						$instr |= ($addr + $offset) % 256;
					} elsif ($token =~ /^\.\-/) {
						$offset = substr($token,2);
						# FixFixFix - Verify decimal number
						$instr |= ($addr - $offset) % 256;
					} else {
						die "Line $linenum: Invalid relative address: $token";
					}
				} elsif (exists $sym_tab{$token}) {
					# A label!
					$offset = ($sym_tab{$token} % 256);
					$instr |= $offset;
				} else {
					# Hex or invalid
					# FixFixFix - interpret hex value
					die "Line $linenum: Invalid target address: $token";
				}
			}
		} elsif (exists $instracir{$token}) {
			# Instructions that reference ac or indirect register
			die "Line $linenum: No snz target specified" if (length($token) eq length($line));
			$token = substr($line,length($token)+1);
			die"Line $linenum: Invalid snz target: $token" unless exists $snztgt{$token};
			$instr = $snztgt{$token};
		} elsif (exists $instrcall{$token}) {
			# CALL instruction
			die "Line $linenum: No call target specified" if (length($token) eq length($line));
			$token = substr($line,length($token)+1);
			$cend = index($token, ',');
			die "Line $linenum: Invalid call target: $token" unless ($cend > 0);
			# strip out white space around comma
			$token =~ s/\)\s+,\s+/\),/g;
			die "Line $linenum: Invalid call target: $token" unless (exists $calltgt{$token});
			$instr = $calltgt{$token};
		} elsif (exists $instrmov{$token}) {
			# MOV instruction
			die "Line $linenum: No mov target specified" if (length($token) eq length($line));
			$token = substr($line,length($token)+1);
			$cend = index($token, ',');
			die "Line $linenum: Invalid mov target: $token" unless ($cend > 0);
			# strip out white space around comma
			$token =~ s/\)\s+,\s+/\),/g;
			die "Line $linenum: Invalid mov target: $token" unless (exists $movtgt{$token});
			$instr = $movtgt{$token};
		} else {
			print "$line\n";
			die "Line $linenum: Unrecognized operation: $token";
		}

		writebin($instr);
		printf {$listfh} "%04d %03X %03X    %s",$linenum,$addr,$instr,$_;
		$addr++;
	}
	

	printf {$listfh} "\n\n\n\nSymbol Table:\n";
	foreach my $symbol (sort keys %main::sym_tab) {
		printf {$listfh} "%-16s %03X\n",$symbol,$main::sym_tab{$symbol};
	}

	close $in or die "$in: $!";
	close $listfh or die "$listfh: $!";
	closebinfile();

	my $end = Time::HiRes::gettimeofday();
	printf "Elapsed time: %.3f\n", $end-$start;
}



# Open the binary file for output
sub openbinfile {
	my $binfilename;
	
	$binfilename = $progname . '.bin';
	open($binfh, ">", $binfilename) or die "Can't open hex file: $binfilename: $!";
	binmode $binfh;
	# Write header: program name followed by a null byte,
	#               followed by Program Start byte (FF)
	syswrite $binfh,$progname,length($progname);
	my $c = chr(0);
	syswrite $binfh,$c,1;
	$c = chr(0xFF);
	syswrite $binfh,$c,1;

	return;
}


# Write a 2-byte record to the binary file
sub writebin {
	my ($rec) = @_;
	my $c = chr(($rec/256) & 0xFF);
	syswrite $binfh,$c,1;
	$c = chr($rec & 0xFF);
	syswrite $binfh,$c,1;

	return;
}


# Close the binary file
sub closebinfile {
	close $binfh  or die "$binfh: $!";
	return;
}



# Trim leading and trailing white space from a string
sub trim {
	my $s = shift;
	$s =~ s/^\s+|\s+$//g;
	return $s;
}


# TO DO:
# Add inc (FF5) & dec (FF6) instructions
# Allow space after comma in mov & call
# Allow hex value as <target>
# Add .blk directive
# Allow '<char> as constant
# Clean up - make subroutine for parsing <target>
